Require Import PArith.
Require Import ZArith.

Section ALIGNDIVIDE.

Open Scope positive_scope.

Lemma mult_leq_r (a b : positive) :
  b <= a * b.
Proof.
  rewrite (eq_refl b : b = 1 * b) at 1.
  unfold Pos.le.
  rewrite Pos.mul_compare_mono_r.
  destruct a; unfold Pos.compare, Pos.compare_cont; discriminate.
Qed.

Lemma mult_leq_l (a b : positive) :
  a <= a * b.
Proof.
  rewrite (Pos.mul_comm a b).
  apply mult_leq_r.
Qed.

Lemma div_leq (m n : positive) :
  (m | n) -> m <= n.
Proof.
  intro H; destruct H; rewrite H.
  apply mult_leq_r.
Qed.

Lemma case_leq_8 (a : positive) :
  a <= 8 -> a = 1 \/ a = 2 \/ a = 3 \/ a = 4 \/ a = 5 \/ a = 6 \/ a = 7 \/ a = 8.
Proof.
  intro H; case_eq a;
    intro a1; case_eq a1;
      intro a2; case_eq a2;
        intro a3; case_eq a3; auto 100;
          first [ intros p _ _ _ Ha | intros _ _ _ Ha];
          rewrite Ha in H;
          contradiction H;
          unfold Pos.le, Pos.compare, Pos.compare_cont;
          try destruct p; trivial.
Qed.

Lemma not_8_divisors : ~(3 | 8) /\ ~(5 | 8) /\ ~(6 | 8) /\ ~(7 | 8).
Proof.
  repeat split;
  unfold Pos.divide;
  intro H; destruct H;
  assert (x <= 8); try (rewrite H; apply mult_leq_l);
  decompose [or] (case_leq_8 x H0);
    first [ rewrite H1 in H | rewrite H2 in H]; discriminate H; contradict H.
Qed.

Lemma case_div_8 (a : positive) :
  (a | 8) -> a = 1 \/ a = 2 \/ a = 4 \/ a = 8.
Proof.
  intro H.
  decompose [and] not_8_divisors.
  decompose [or] (case_leq_8 a (div_leq a 8 H));
    [ auto  (* 1 *)
    | auto  (* 2 *)
    | rewrite H3 in H; contradict (H0 H)
    | auto  (* 4 *)
    | rewrite H3 in H; contradict (H2 H)
    | rewrite H5 in H; contradict (H1 H)
    | rewrite H3 in H; contradict (H4 H)
    | auto ].
Qed.

Let P (a a' : positive) :
  a <= a' -> (a | 8) -> (a' | 8) -> (a | a').
Proof.
  intros leq a_div_8 a'_div_8.
  unfold Pos.divide.

  decompose [or] (case_div_8 a a_div_8); decompose [or] (case_div_8 a' a'_div_8);
  try rewrite H in *; try rewrite H0 in *; try rewrite H1 in *;

  solve [ exists 1; trivial
        | exists 2; trivial
        | exists 4; trivial
        | exists 8; trivial
        | unfold Pos.le, Pos.compare, Pos.compare_cont in leq; contradict leq; trivial].
Qed.

Open Scope Z_scope.

Let L (a a' : Z) :
  0 < a <= a' -> (a | 8) -> (a' | 8) -> (a | a').
Proof.
  intros order a_div_8 a'_div_8.
  destruct order as [ a_gt_0 a_le_a' ].
  assert (0 < a') as a'_gt_0.
    apply Z.lt_le_trans with (m := a); assumption.
  case_eq a.
    intro a_eq_0; rewrite a_eq_0 in a_gt_0.
    unfold Z.lt, Z.compare in a_gt_0; discriminate a_gt_0.

    case_eq a'.
      intro a'_eq_0; rewrite a'_eq_0 in a'_gt_0.
      unfold Z.lt, Z.compare in a'_gt_0; discriminate a'_gt_0.

      (* core case *)
      intros p' pa' p pa; rewrite pa, pa' in *; apply Pos2Z.inj_divide; apply P;
      solve [ unfold Z.le, Z.compare in a_le_a'; assumption
            | apply Pos2Z.inj_divide; assumption ].
   
      intros p' pa'.
      rewrite pa' in a'_gt_0; unfold Z.lt, Z.compare in a'_gt_0; discriminate a'_gt_0.

    intros p pa.
    rewrite pa in a_gt_0; unfold Z.lt, Z.compare in a_gt_0; discriminate a_gt_0.
Qed.

Lemma align_divides_8:
  forall al,
    0 < al ->
    (al | 8) ->
    {al = 1} + {al = 2} + {al = 4} + {al = 8}.
Proof.  
  intros. destruct (Z_eq_dec al 1); auto. destruct (Z_eq_dec al 2); auto. destruct (Z_eq_dec al 4); auto. destruct (Z_eq_dec al 8); auto. exfalso.
  destruct al; try discriminate.
  apply Pos2Z.inj_divide in H0.
  apply case_div_8 in H0.
  destruct H0; try congruence.
  destruct H0; try congruence.
  destruct H0; congruence.
Defined.

Lemma divides_8_le_divides: forall al1 al,
 0 < al1 <= al ->  (al | 8) -> (al1 | 8) -> forall sz, (al | sz) -> (al1 | sz).
Proof.
  intros. eapply Z.divide_trans. eapply L. eassumption. assumption. assumption. assumption.
Qed.

End ALIGNDIVIDE.
