Require Import Coqlib.
Require Import FSets.
Require FSetAVL.
Require Import Orders.
Require Mergesort.
Require Import AST.
Require Import Csharpminor.
Require Import Tagmap.
Require Import Maps.
Require Import Permutation.
Require Import AlignDivide.
Require Import StrongInject.

Require Import Errors.

Module VarOrder <: TotalLeBool.
  Definition t := (ident * (Z * Z * option Z))%type.
  Definition leb (v1 v2: t) : bool := 
    match (v1, v2) with
      | ((_, (_, al1, _)), (_, (_, al2, _))) => zle al2 al1
    end.
  Theorem leb_total: forall v1 v2, leb v1 v2 = true \/ leb v2 v1 = true.
  Proof.
    unfold leb; intros. 
    destruct v1. destruct p. destruct p.
    destruct v2. destruct p. destruct p.
    assert (z2 <= z0 \/ z0 <= z2) by omega.
    unfold proj_sumbool. destruct H; [left|right]; apply zle_true; auto.
  Qed.
End VarOrder.

Module VarSort := Mergesort.Sort(VarOrder).

Function compute_offsets_aux (of: Z) (accu: PTree.t Z) (l: list (ident * (Z * Z * option Z))) {struct l}: PTree.t Z :=
  match l with
    | nil => accu
    | (id, (sz, al, _)) :: l' => compute_offsets_aux (of + sz) (PTree.set id of accu) l'
  end.

Lemma compute_offsets_aux_accu_exists:
  forall accu id,
    accu ! id <> None ->
    forall of l, (compute_offsets_aux of accu l) ! id <> None.
Proof.
  intros. revert accu H of. induction l; simpl.
   tauto.
  destruct a. destruct p. destruct p. intros.
  eapply IHl.
  destruct (peq id i).
   subst. rewrite PTree.gss. congruence.
  rewrite PTree.gso; auto.
Qed.

Lemma compute_offsets_aux_exists:
  forall l accu id,
    (accu ! id <> None \/ (exists szal, In (id, szal) l)) ->
    forall of, (compute_offsets_aux of accu l) ! id <> None.
Proof.
  induction l; simpl.
   destruct 1. auto. destruct H. contradiction.
  destruct 1; subst.
   destruct a. destruct p0. destruct p0. intros. eapply compute_offsets_aux_accu_exists.
   destruct (peq id p).
    subst. rewrite PTree.gss. congruence.
   rewrite PTree.gso; auto.
  destruct a. destruct p0. destruct p0.
  destruct H.
  destruct H.
   inv H. intros. eapply compute_offsets_aux_accu_exists. rewrite PTree.gss. congruence.
  eauto.
Qed.  

Lemma compute_offsets_aux_not_in:
  forall l id,
    ~ In id (map fst l) ->
    forall of accu,
      (compute_offsets_aux of accu l) ! id = accu ! id.
Proof.
  induction l; simpl.
   reflexivity.
  destruct a; simpl.
  destruct p0. destruct p0.
  intros.
  rewrite IHl.
  apply PTree.gso.
  cut (p <> id). auto. tauto.
  tauto.
Qed.

Lemma varorder_hdrel:
  forall l id sz al o,
    HdRel (fun a b => is_true (VarOrder.leb a b)) (id, (sz, al, o)) l ->
    Sorted (fun a b => is_true (VarOrder.leb a b)) l ->
    forall id' sz' al' o',
      In (id', (sz', al', o')) l ->
      al' <= al.
Proof.
  induction l; inversion 1; subst.
   simpl; tauto.
  inversion 1; subst.
  inversion 1; subst.
   unfold is_true, VarOrder.leb in H1.
   destruct (zle al' al).
    omega.
   discriminate.
  destruct a. destruct p. destruct p.
  unfold is_true, VarOrder.leb in H1.
  destruct (zle z0 al).
   2: discriminate.
  refine (_ (IHl _ _ _ _ _ _ _ _ _ _ _)).
  2: eassumption.
  2: assumption.
  2: eassumption.
  intro.
  omega.
Qed.

Lemma stack_size_pos:
  forall l,
    (forall id sz al o, In (id, (sz, al, o)) l -> 0 <= sz) ->
    0 <= stack_size l.
Proof.
  induction l; simpl.
   omega.
  intros.
  refine (_ (IHl _)).
  intro. destruct a. destruct p. destruct p.
  generalize (H _ _ _ _ (or_introl _ (refl_equal _))).
  omega.
  eauto.
Qed. 

Theorem compute_offsets_aux_correct:
forall l,
  Sorted (fun a b => is_true (VarOrder.leb a b)) l ->
  list_norepet (map fst l) ->
  forall of,
  (forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8) /\ (al | of))) ->
  forall id sz al o, 
    In (id, (sz, al, o)) l ->
    0 <= of ->
    forall accu,  
    exists of', PTree.get id (compute_offsets_aux of accu l) = Some of' /\
                of <= of' /\ of' + sz <= of + stack_size l /\
                (al | of')
.
Proof.
  intros.
  revert H H0 H1 id sz al o H2 H3.
  functional induction (compute_offsets_aux of accu l); simpl.
   tauto.
  inversion 1; subst.
  inversion 1; subst.
  destruct 2.
   inv H4. intros. rewrite compute_offsets_aux_not_in. rewrite PTree.gss.
   exploit H1.
   left. reflexivity.
   destruct 1. destruct H8. destruct H9. destruct H10.
   esplit. split. reflexivity. split. omega. split.
    assert (0 <= stack_size l'). apply stack_size_pos. intros. exploit H1. right. eassumption. tauto.
    omega.
    assumption.
  assumption.
 intros.
 exploit IHt.
 assumption.
 assumption.
 intros.
 exploit H1.
 right. eassumption.
 destruct 1. destruct H10. destruct H11. destruct H12.
 split; auto. split; auto. split; auto. split; auto.
 exploit H1. left; reflexivity. destruct 1. destruct H15. destruct H16. destruct H17.
 assert (al1 <= al) by eauto using varorder_hdrel.
 apply Z.divide_add_r.
 assumption.
 eauto using divides_8_le_divides.
 eassumption.
 exploit H1.
 left. reflexivity.
 omega.
 destruct 1. destruct H8. destruct H9. destruct H10.
 esplit. split. eassumption. split.
 exploit H1. left. reflexivity. destruct 1. destruct H12. omega.
 split. omega.
 assumption.
Qed.

Theorem compute_offsets_aux_disjoint_aux:
  forall l1 l2,
    Sorted (fun a b => is_true (VarOrder.leb a b)) (l1 ++ l2) ->
    list_norepet (map fst (l1 ++ l2)) ->
    forall of,
      (forall id sz al o, In (id, (sz, al, o)) (l1 ++ l2) -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8) /\ (al | of))) ->
      0 <= of ->
      forall id1 sz1 al1 o1, 
        In (id1, (sz1, al1, o1)) l1 ->
        forall accu of1,
          PTree.get id1 (compute_offsets_aux of accu (l1 ++ l2)) = Some of1 ->
          forall id2 sz2 al2 o2,
            In (id2, (sz2, al2, o2)) l2 ->
            forall of2,
              PTree.get id2 (compute_offsets_aux of accu (l1 ++ l2)) = Some of2 ->
              of1 + sz1 <= of2.
Proof.
  induction l1; simpl.
   tauto.
  inversion 1; subst.
  inversion 1; subst.
  destruct a.
  destruct p.
  destruct p.
  intros until 2.
   assert (
        forall (id : ident) (sz al : Z) (o : option Z),
          In (id, (sz, al, o)) (l1 ++ l2) ->
       0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8) /\ (al | of + z)
     ).
    intros. exploit H1. right. eassumption.
    destruct 1. destruct H9. destruct H10. destruct H11. split; auto. split; auto. split; auto. split; auto.
    apply Z.divide_add_r.
    assumption.
    eapply divides_8_le_divides. split. assumption. eapply varorder_hdrel. eassumption. assumption. eassumption. exploit H1. left. reflexivity. tauto. assumption.
   exploit H1. left. reflexivity. tauto.
  destruct 1.
   inv H8.
   intro.
   rewrite compute_offsets_aux_not_in.
   rewrite PTree.gss.
   intros until 1. inv H8.
   intros until 1.
   exploit compute_offsets_aux_correct.   
   eexact H2.
   assumption.
   eassumption.
   eapply in_or_app. right. eassumption.
   exploit H1. left. reflexivity. omega.
   instantiate (1 := PTree.set id1 of1 accu).
   destruct 1.
   destruct H9. destruct H10. destruct H11.
   rewrite H9.
   intros until 1. inv H13. tauto.
   assumption.
  intros.
  eapply IHl1.
  eassumption.
  assumption.
  eassumption.
  exploit H1. left. reflexivity. omega.
  eassumption.
  eassumption.
  eassumption.
  assumption.
Qed.

Theorem compute_offsets_aux_disjoint:
  forall l,
    Sorted (fun a b => is_true (VarOrder.leb a b)) l ->
    list_norepet (map fst l) ->
    forall of,
      (forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8) /\ (al | of))) ->
      0 <= of ->
      forall id1 sz1 al1 o1, 
        In (id1, (sz1, al1, o1)) l ->
        forall accu of1,
          PTree.get id1 (compute_offsets_aux of accu l) = Some of1 ->
          forall id2 sz2 al2 o2,
            In (id2, (sz2, al2, o2)) l ->
            forall of2,
              PTree.get id2 (compute_offsets_aux of accu l) = Some of2 ->
              id1 <> id2 ->
              of1 + sz1 <= of2 \/ of2 + sz2 <= of1.
Proof.
  intros.
  exploit in_split.
  eexact H3.
  destruct 1. destruct H8. subst.
  destruct (in_app_or _ _ _ H5).
   right. eapply compute_offsets_aux_disjoint_aux.
   eassumption.
   assumption.
   eassumption.
   assumption.
   eassumption.
   eassumption.
   left. reflexivity.
   assumption.
  inversion H8; subst.
   inv H9. destruct H7. reflexivity.
  assert (In (id1, (sz1, al1, o1)) (x ++ (id1, (sz1, al1, o1)) :: nil)).
   apply in_or_app.
   right. left. reflexivity.
  assert (x ++ (id1, (sz1, al1, o1)) :: x0 = (x ++ (id1, (sz1, al1, o1)) :: nil) ++ x0).
   rewrite app_ass. reflexivity.
  unfold VarOrder.t, ident in *. rewrite H11 in *.
  left.
  eapply compute_offsets_aux_disjoint_aux.
  eassumption.
  assumption.
  eassumption.
  assumption.
  eassumption.
  eassumption.
  eassumption.
  assumption.
Qed.

Definition compute_offsets l := compute_offsets_aux 0 (PTree.empty _) (VarSort.sort l).

Lemma compute_offsets_exists:
  forall l id szal, In (id, szal) l ->
                    (compute_offsets l) ! id <> None.
Proof.
  intros. eapply compute_offsets_aux_exists.
  right.
  exists szal.
  eapply Permutation_in.
  eapply VarSort.Permuted_sort.
  assumption.
Qed.

Lemma stack_size_permut:
  forall l1 l2,
    Permutation l1 l2 ->
    stack_size l1 = stack_size l2.
Proof.
  induction 1; simpl.
   reflexivity.
  rewrite IHPermutation. reflexivity.
  destruct x. destruct p. destruct p. destruct y. destruct p. destruct p. omega.
  congruence.
Qed.

Lemma list_norepet_permutation:
  forall {A: Type} (l l': _ A), Permutation l l' -> list_norepet l -> list_norepet l'.
Proof.
  induction 1.
   tauto.
  inversion 1; subst. constructor. 
  intro. destruct H3. apply Permutation_sym in H. eauto using  Permutation_in.
  auto.
  inversion 1; subst. inversion H3; subst.
  constructor.
   intro. inversion H0; subst; contradiction.
  constructor. intro. unfold In in *. tauto.
  assumption.
  auto.
Qed.

Theorem compute_offsets_correct:
forall l,
  list_norepet (map fst l) ->
  (forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))) ->
  forall id sz al o, 
    In (id, (sz, al, o)) l ->
    exists of', PTree.get id (compute_offsets l) = Some of' /\
                0 <= of' /\ of' + sz <= stack_size l /\
                (al | of')
.
Proof.
  intros. 
  replace (stack_size l) with (0 + stack_size l) by omega.
  erewrite stack_size_permut.
  2: eapply VarSort.Permuted_sort.
  eapply compute_offsets_aux_correct.
  eapply VarSort.LocallySorted_sort.
  eapply list_norepet_permutation.
  eapply Permutation_map.
  eapply VarSort.Permuted_sort.
  assumption.
 intros.
 exploit H0.
  eapply Permutation_in. eapply Permutation_sym. eapply VarSort.Permuted_sort. eassumption.
  destruct 1.  destruct H4. destruct H5.
  split; auto. split; auto. split; auto. split; auto. exists 0; auto.
  eapply Permutation_in. eapply VarSort.Permuted_sort. eassumption.
  omega.
Qed.

Theorem compute_offsets_disjoint:
  forall l,
    list_norepet (map fst l) ->
      (forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))) ->
      forall id1 sz1 al1 o1, 
        In (id1, (sz1, al1, o1)) l ->
        forall of1,
          PTree.get id1 (compute_offsets l) = Some of1 ->
          forall id2 sz2 al2 o2,
            In (id2, (sz2, al2, o2)) l ->
            forall of2,
              PTree.get id2 (compute_offsets l) = Some of2 ->
              id1 <> id2 ->
              of1 + sz1 <= of2 \/ of2 + sz2 <= of1.
Proof.
  intros.
  eapply compute_offsets_aux_disjoint.
  eapply VarSort.LocallySorted_sort.
  eapply list_norepet_permutation.
  eapply Permutation_map. eapply VarSort.Permuted_sort. eassumption.
  instantiate (1 := 0). intros. exploit H0.
  eapply Permutation_in. eapply Permutation_sym. eapply VarSort.Permuted_sort. eassumption.
  destruct 1. destruct H8. destruct H9. split; auto. split; auto. split; auto. split; auto. exists 0; omega.
  omega.
  eapply Permutation_in. eapply VarSort.Permuted_sort. eassumption.
  eassumption.
  eapply Permutation_in. eapply VarSort.Permuted_sort. eassumption.
  assumption.
  assumption.
Qed.

Definition compile_vars' (t: PTree.t Z) : list (_ * (Z * Z * option Z))%type -> _ := List.map 
                               (fun isao => match isao with
                                              | (id, (sz, al, _)) => (id, (sz, al, PTree.get id t))
                                            end).

Lemma check_offsets_exists_aux: forall l t,
(forall id sza, In (id, sza) l -> t ! id <> None) ->
forall accu,
exists l', check_offsets accu (compile_vars' t l) = OK l'.
Proof.
  induction l; simpl.
   eauto.
  intros. destruct a. destruct p0. destruct p0.
  case_eq (t ! p).
   eauto.
  intro. edestruct H. left. reflexivity. assumption.
Qed.

Definition compile_vars l := compile_vars' (compute_offsets l) l.

Lemma check_offsets_exists: forall l,
exists l', check_offsets nil (compile_vars l) = OK l'.
Proof.
  intros. eapply check_offsets_exists_aux.
  destruct sza. destruct p. intros.
  eapply compute_offsets_exists.
  eassumption.
Qed.

Lemma stack_size_map: forall l,
                        stack_size
                          (map
                             (fun isao : positive * (Z * Z * option Z) =>
                                let (id, y) := isao in
                                let (y0, _) := y in
                                let (sz, al) := y0 in (id, (sz, al, (compute_offsets l) ! id))) l)
                        = stack_size l.
Proof.
  intro.
  generalize (compute_offsets l).
  induction l; simpl.
  reflexivity.
  destruct a. destruct p. destruct p. intro. erewrite IHl. reflexivity.
Qed.

Lemma assign_variables_aux_map: forall l t b,
                                  assign_variables_aux t b (map
                                                              (fun isao : positive * (Z * Z * option Z) =>
                                                                 let (id, y) := isao in
                                                                 let (y0, _) := y in
                                                                 let (sz, al) := y0 in (id, (sz, al, (compute_offsets l) ! id))) l) = assign_variables_aux t b l.
Proof.
  intro. generalize (compute_offsets l). induction l; simpl.
   trivial.
  intros. destruct a. destruct p. destruct p.
  auto.
Qed.
                                         
Theorem call_condition_ok:
  forall l,
    list_norepet (map fst l) ->
    (forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))) ->
    call_condition true (compile_vars l).
Proof.
  unfold call_condition, compile_vars, compile_vars'.
  split.
   intros. destruct (list_in_map_inv _ _ _ H1). destruct H2. destruct x. destruct p0. destruct p0. inv H2. exploit compute_offsets_correct. eassumption. assumption. eassumption. destruct 1. destruct H2. destruct H4. destruct H5. eauto.
  split.
   intros. destruct (list_in_map_inv _ _ _ H1). destruct H4. destruct x. destruct p0. destruct p0. inv H4.
   destruct (list_in_map_inv _ _ _ H2). destruct H4. destruct x. destruct p1. destruct p1. inv H4.
   eapply compute_offsets_disjoint.
   eassumption.
   assumption.
   eassumption.
   congruence.
   eassumption.
   congruence.
   assumption.
  intros.  destruct (list_in_map_inv _ _ _ H1). destruct H2. destruct x. destruct p0. destruct p0. inv H2.
  exploit compute_offsets_correct. eassumption. assumption. eassumption. destruct 1. destruct H2. rewrite H2 in H8. inv H8. destruct H4. destruct H5. rewrite stack_size_map. auto.
Qed.

Lemma list_norepet_compile_vars:
  forall l,
    list_norepet (map fst l) ->
    list_norepet (map fst (compile_vars l)).
Proof.
  unfold compile_vars, compile_vars'. intro. generalize (compute_offsets l). intro.
  induction l; simpl; auto.
  inversion 1; subst.
  destruct a. destruct p. destruct p. simpl in *.
  constructor.
   intro. destruct (list_in_map_inv _ _ _ H0). destruct H1; subst. destruct x; simpl.
   destruct (list_in_map_inv _ _ _ H4).
   destruct H1. destruct x. destruct p2. destruct p2. inv H1.
   simpl in *.
   generalize (in_map fst _ _ H5). assumption.
  auto.
Qed.
   

Section MEM.

(*
Lemma check_alignment_size_cond sz al : {0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8)} + {~ (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))}.
Proof.
  destruct (zlt 0 al); try (right; omega).
  destruct (zle 0 sz); try (right; omega).
  assert (al > 0) by omega.
  destruct (Zdivide_dec al sz H); try (right; tauto).
  destruct (Zdivide_dec al 8 H); tauto.
Defined.

Lemma check_alignment_size: forall l: list (ident * (Z * Z * option Z)), {forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))} + {~ forall id sz al o, In (id, (sz, al, o)) l -> (0 < al /\ 0 <= sz /\ (al | sz) /\ (al | 8))}.
Proof.
  induction l; simpl.
   left. tauto.
  destruct a. destruct p. destruct p.
  destruct (check_alignment_size_cond z z0).
   destruct IHl. left. destruct 1; eauto. congruence.
  right. intro. edestruct n. eauto.
 right. intro. edestruct n. eauto.
Defined.
*)

(*
Function transf_function (f: function) : res function :=
  if list_norepet_dec peq (map fst f.(fn_vars))
  then
    if check_alignment_size f.(fn_vars)
    then
      OK (mkfunction f.(fn_id) f.(fn_sig) f.(fn_params) (compile_vars f.(fn_vars)) f.(fn_temps) f.(fn_body))
    else Error (msg "Invalid alignments and sizes")
  else Error (msg "Duplicate variables").
*)

Definition transf_function f :=
  mkfunction f.(fn_id) f.(fn_sig) f.(fn_params) (compile_vars f.(fn_vars)) f.(fn_temps) f.(fn_body).

Inductive match_cont: cont -> cont -> Prop :=
  | match_cont_stop: match_cont Kstop Kstop
  | match_cont_seq: forall k1 k2 (HCONT_IND: match_cont k1 k2) s, match_cont (Kseq s k1) (Kseq s k2)
  | match_cont_block: forall k1 k2 (HCONT_IND: match_cont k1 k2), match_cont (Kblock k1) (Kblock k2)
  | match_cont_call: forall f1 f2
                            (HCONT_FUNC: transf_function f1 = f2)
                       k1 k2
                       (HCONT_IND: match_cont k1 k2)
                       b o e t, 
                         match_cont (Kcall b o f1 e t k1) (Kcall b o f2 e t k2).

Lemma match_call_cont: forall k k2, match_cont k k2 -> match_cont (call_cont k) (call_cont k2).
Proof.
  induction 1; simpl; auto; constructor; auto.
Qed.

Scheme stmt_ind2 := Induction for stmt Sort Prop
                    with lbl_stmt_ind2 := Induction for lbl_stmt Sort Prop.

Combined Scheme stmt_lbl_stmt_ind2 from stmt_ind2, lbl_stmt_ind2.

Lemma match_find_label_aux:
  (forall s lbl k,
    match find_label lbl s (k) with
        | Some (s', k') =>
          forall k2, match_cont k k2 -> exists k'2, match_cont k' k'2 /\ find_label lbl s (k2) = Some (s', k'2)
        | None => forall k2, find_label lbl s k2 = None
    end)
    /\
    (forall l (lbl : label) (k : cont),
         match find_label_ls lbl l k with
           | Some (s', k') =>
             forall k2 : cont,
               match_cont k k2 ->
               exists k'2 : cont,
                 match_cont k' k'2 /\ find_label_ls lbl l k2 = Some (s', k'2)
           | None => forall k2 : cont, find_label_ls lbl l k2 = None
         end).
Proof.
  apply stmt_lbl_stmt_ind2; simpl; auto.

  (* seq *)
  intros. generalize (H lbl (Kseq s0 k)); clear H.
  case_eq (find_label lbl s (Kseq s0 k)).
   destruct p. intros.
   exploit H1; clear H1. constructor. eassumption.
   destruct 1. destruct H1. rewrite H3. eauto.
  intros. generalize (H0 lbl k); clear H0.
  case_eq (find_label lbl s0 k).
   destruct p. intros. rewrite H1. auto.
  intros. rewrite H1. auto.

  (* if?? *)
  intros.
  generalize (H lbl k); clear H.
  case_eq (find_label lbl s k).
   destruct p. intros.
   exploit H1; clear H1. eassumption. destruct 1. destruct H1. rewrite H3.
   eauto.
  intros.
  generalize (H0 lbl k); clear H0.
  case_eq (find_label lbl s0 k).
  destruct p. intros.
  rewrite H1. auto.
  intros. rewrite H1. auto.

  (* seq loop *)
  intros.
  generalize (H  lbl (Kseq (Sloop s) k)); clear H.
  case_eq (find_label lbl s (Kseq (Sloop s) k)); auto.
  destruct p. intros.
  eapply H0. constructor. eassumption.

  (* block *)
  intros.
  generalize (H lbl (Kblock k)); clear H.
  case_eq (find_label lbl s (Kblock k)); auto.
  destruct p. intros.
  eapply H0. constructor. assumption.

  (* ls *)
  intros. destruct (ident_eq lbl l).
   subst. eauto.
  apply H.

  (* switch *)
  intros.
  generalize (H lbl (Kseq (seq_of_lbl_stmt l) k)); clear H.
  case_eq (find_label lbl s (Kseq (seq_of_lbl_stmt l) k)).
   destruct p. intros. exploit H1; clear H1. econstructor; eauto.
   destruct 1. destruct H1. rewrite H3. eauto.
  intros. generalize (H0 lbl k); clear H0.
  case_eq (find_label_ls lbl l k).
   destruct p. intros. rewrite H1. auto.
  intros. rewrite H1. auto.
Qed.
  
Require Import Globalenvs.

Lemma function_id_correct (p: program):  
  {forall b f,
    Genv.find_funct_ptr (Genv.globalenv p) b = Some (Internal f) ->
    forall fid, Genv.find_symbol (Genv.globalenv p) fid = Some b ->
                f.(fn_id) = fid} + {~forall b f,
    Genv.find_funct_ptr (Genv.globalenv p) b = Some (Internal f) ->
    forall fid, Genv.find_symbol (Genv.globalenv p) fid = Some b ->
    f.(fn_id) = fid}.
Proof.
  cut (
      forall n, 0 <= n -> {forall b, 0 < b < n -> forall f,
    Genv.find_funct_ptr (Genv.globalenv p) b = Some (Internal f) ->
    forall fid, Genv.find_symbol (Genv.globalenv p) fid = Some b ->
                f.(fn_id) = fid} + {~forall b, 0 < b < n -> forall f,
    Genv.find_funct_ptr (Genv.globalenv p) b = Some (Internal f) ->
    forall fid, Genv.find_symbol (Genv.globalenv p) fid = Some b ->
    f.(fn_id) = fid}
    ).
   intros.
   refine (_ (H (Genv.genv_next (Genv.globalenv p)) _)); clear H.
   destruct 1.
    left. intros.
     eapply e; eauto. unfold Genv.find_funct_ptr in H. eauto using Genv.genv_funs_range.
    right. intro.
    destruct n.
    eauto.
    generalize (Genv.genv_next_pos (Genv.globalenv p)). omega.
   apply natlike_rec.
    left. intros; omega.
   intros.
   destruct H0.
    case_eq (Genv.find_funct_ptr (Genv.globalenv p) x).
     destruct f.
     intros.
     case_eq (Genv.invert_symbol (Genv.globalenv p) x).
      intros.
      generalize (Genv.invert_find_symbol _ _ H1).
      intro.
      destruct (peq (fn_id f) i).
       subst.
       left.
       intros.
       destruct (zeq b x).
        subst. replace f0 with f in * by congruence. eapply Genv.genv_vars_inj. eassumption. assumption.
       eapply e. 2: eassumption. omega. assumption.
      right.
      intro.
      apply n.
      eapply H3. 2: eassumption.
      2: assumption.
      unfold Genv.find_symbol in H2.
      generalize (Genv.genv_symb_range _ _ H2).
      omega.
     intros.
     left.
     intros.
     destruct (zeq b x).
      subst.
      replace f0 with f in * by congruence.
      generalize (Genv.find_invert_symbol _ _ H4). intros; congruence.
     eapply e. 2: eassumption. omega. assumption.
    intros. left. intros.
    destruct (zeq b x).
     subst. congruence.
    eapply e. 2: eassumption. omega. assumption.
   intros. left. intros.
   destruct (zeq b x).
   subst. congruence.
  eapply e. 2: eassumption. omega. assumption.
 right. intro. eapply n. intros. eapply H0; eauto. omega.
Defined.

Function transf_program (p: program): res program :=
  if function_id_correct p
  then OK (AST.transform_program (transf_fundef transf_function) p)
  else Error (msg "Function identifiers do not match. This should not happen").


Require Import Memory.
Variable p: program.
Context `{M: Mem.MEM}.

Function function_is_defined (f: ident) : option (option function) :=
  match Genv.find_symbol (Genv.globalenv p) f with
    | None => Some None
    | Some b =>
      match Genv.find_funct_ptr (Genv.globalenv p) b with
        | Some (Internal f) => Some (Some f)
        | Some (External _) => Some None
        | None =>
          match Genv.find_var_info (Genv.globalenv p) b with
            | Some _ => None
            | None => Some None
          end
      end
  end.

Variable stfun_other: Mem.stag -> option Mem.stag.

Function stfun (t: Mem.stag) {struct t}: option Mem.stag :=
  match t with
    | Mem.Stack_main fi os => 
      match function_is_defined fi with
        | Some (Some f) => 
          match os with
            | Some _ => None                          
            | None => Some (Mem.Stack_main fi (Some (stack_size f.(fn_vars))))
          end
        | Some None => stfun_other t
        | None => None
      end
    | Mem.Stack_var (Mem.Svar fi v b sz al os) =>
      match function_is_defined fi with
        | Some (Some f) =>
          match os with
            | Some _ => None
            | None =>
              match (compute_offsets f.(fn_vars)) ! v with
                | Some o => Some (Mem.Stack_var (Mem.Svar fi v b sz al (Some o)))
                | None => None
              end
          end
        | Some None =>  stfun_other t
        | None => None
      end
  end.

Context `{TM: Mem.TAGMAP _ (M := M) stfun}.

Inductive match_states: state -> state -> Prop :=
| match_states_state: forall f1 f2
                             (HFUN: transf_function f1 = f2)
                             k1 k2
                        (HCONT: match_cont k1 k2)
                          m1 m2
                                 (HMEM: Mem.tagmap stfun m1 = Some m2)
                                 b s e le,
                              match_states (State b f1 s k1 e le m1) (State b f2 s k2 e le m2)

| match_states_callstate: forall f1 f2
                            (HFUN: transf_fundef transf_function f1 =f2)
                              k1 k2 (HCONT: match_cont k1 k2)
                                           m1 m2
  (HMEM: Mem.tagmap stfun m1 = Some m2)
 args,
                                              match_states (Callstate f1 args k1 m1) (Callstate f2 args k2 m2)

| match_states_returnstate: forall k1 k2 (HCONT: match_cont k1 k2)
                                          m1 m2 (HMEM: Mem.tagmap stfun m1 = Some m2)
                                                         v,
                                                          match_states (Returnstate v k1 m1) (Returnstate v k2 m2)
.                         

  Lemma tagmap_region_free_aux:
    forall b m l m',
      region_free_aux b m l = Some m' ->
      forall tm,
        Mem.tagmap stfun m = Some tm ->
        exists tm',
          Mem.tagmap stfun m' = Some tm' /\
          region_free_aux b tm l = Some tm'.
  Proof.
    intros until l. functional induction (region_free_aux b m l); simpl; try discriminate.
     intros until 1. inv H; eauto.
    intros.
    exploit Mem.tagmap_free.
    eassumption.
    eassumption.
    destruct 1. destruct H1. rewrite H2. eauto.
  Qed.

  Lemma compile_vars_region_free_aux: forall l b m,
     region_free_aux b m (map snd (compile_vars l)) = region_free_aux b m (map snd l).
  Proof.
    intro. unfold compile_vars, compile_vars'. generalize (compute_offsets l). induction l; simpl.
     reflexivity.
    destruct a. destruct p0. destruct p0. simpl. intros.
    case_eq (Mem.free m b 0 z); auto.
  Qed.

Lemma tagmap_perm': forall m m', (Mem.tagmap stfun m = Some m' \/ Mem.tagmap stfun m' = Some m) ->
                                 forall b o k p,
                                   Mem.perm m b o k p -> Mem.perm m' b o k p.
Proof.
  destruct 1; intros until p0.
   generalize (Mem.tagmap_perm _ _ _ H b o k p0). tauto.
  generalize (Mem.tagmap_perm _ _ _ H b o k p0). tauto.
Qed.

Lemma tagmap_valid_access: forall m m', (Mem.tagmap stfun m = Some m' \/ Mem.tagmap stfun m' = Some m) ->
                                        forall chunk b o p,
                                          (Mem.valid_access m chunk b o p -> Mem.valid_access m' chunk b o p).
Proof.
  unfold Mem.valid_access, Mem.range_perm;
  destruct 2; split; auto; intros; eapply tagmap_perm'; eauto.
Qed.

Lemma tagmap_valid_pointer: forall m m', Mem.tagmap stfun m = Some m' ->
                                         Mem.valid_pointer m' = Mem.valid_pointer m.
Proof.
  intros. apply FunctionalExtensionality.functional_extensionality.
  intros.
  apply FunctionalExtensionality.functional_extensionality.
  intros.
  case_eq (Mem.valid_pointer m' x x0); case_eq (Mem.valid_pointer m x x0); trivial; intros; exfalso.
   apply Mem.valid_pointer_valid_access in H1.
   eapply tagmap_valid_access in H1. 2: eauto. apply Mem.valid_pointer_valid_access in H1. congruence.
  apply Mem.valid_pointer_valid_access in H0.
  eapply tagmap_valid_access in H0. 2: eauto. apply Mem.valid_pointer_valid_access in H0. congruence.
Qed.

Hypothesis stfun_other_align:
  forall st st',
    stfun_other st = Some st' ->
    Mem.tag_align (Mem.Tag_stack st) = Mem.tag_align (Mem.Tag_stack st').

Lemma tagmap_option_tag_align:
  forall m m', 
    Mem.tagmap stfun m = Some m' ->
    forall b,
    Mem.option_tag_align (Mem.tget m' b) = Mem.option_tag_align (Mem.tget m b).
Proof.
  intros.
  case_eq (Mem.tget m b).
  intros. exploit Mem.tagmap_tget_some. eassumption. eassumption.
  destruct 1. destruct H1. rewrite H2.
  functional inversion H1; subst.
   reflexivity.
   functional inversion H6; subst; try reflexivity.
   Opaque Mem.tag_align. simpl. erewrite <- stfun_other_align. 2: eassumption. reflexivity.
   simpl. erewrite <- stfun_other_align. 2: eassumption. reflexivity.
  intro. erewrite Mem.tagmap_tget_none. reflexivity. eassumption. assumption.  
Qed.

Lemma tagmap_region_alloc:
  forall f bf,
    Genv.find_symbol (Genv.globalenv p) (fn_id f) = Some bf ->
    Genv.find_funct_ptr (Genv.globalenv p) bf = Some (Internal f) ->
    forall l' l'',
      compile_vars' (compute_offsets (fn_vars f)) 
                    (map (fun idsza : ident * (Z * Z) =>
                            let (id, y) := idsza in let (sz, al) := y in (id, (sz, al, None))) l') =
      map
        (fun i : ident * (Z * Z * Z) =>
           let (id, y) := i in
           let (y0, al) := y in let (sz, of) := y0 in (id, (sz, of, Some al)))
        l'' ->
      forall b m1 m2, Mem.tagmap stfun m1 = Some m2 ->
                    Mem.tagmap stfun (region_alloc' b (fn_id f) m1 l') =
                    Some (Mem.region_alloc b (fn_id f) m2 l'').
Proof.
  induction l'; destruct l''; simpl; try congruence.
  destruct a. destruct p1.
  destruct p0. destruct p0. destruct p0.
  intro.
  inv H1.
  intros.
  rewrite H6.
  case_eq (Mem.alloc m1 0 z2 (Mem.Tag_stack (Mem.Stack_var (Mem.Svar (fn_id f) i0 (b) z2 z3 None)))).
   intros.
   exploit Mem.tagmap_alloc.
   eassumption.
   eassumption.
   simpl.
   unfold function_is_defined.
   rewrite H. rewrite H0. rewrite H6. reflexivity.
  destruct 1. destruct H3.
  rewrite H4.
  eauto.
Qed.

Variable tp: program.
Hypothesis TRANSF_PROGRAM: transf_program p = OK tp.

Lemma eval_expr_correct:
  forall e le m a v,
    eval_expr (Genv.globalenv p) e le m a v ->
    forall m',
      Mem.tagmap stfun m = Some m' ->
      eval_expr (Genv.globalenv tp) e le m' a v.
Proof.
  induction 1; try (econstructor; eauto; fail).
  inversion H; subst.
   econstructor. eleft. eassumption.
  econstructor. eright. assumption.
  functional inversion TRANSF_PROGRAM.
  erewrite Genv.find_symbol_transf. eassumption.
  intros. econstructor. eauto. eauto.
  revert H1. destruct op; simpl; try congruence.
  erewrite <- tagmap_valid_pointer. intro. eassumption. assumption.
 intros. econstructor. eauto. erewrite Mem.tagmap_load. eassumption. assumption.
 erewrite tagmap_option_tag_align. eassumption. assumption.
 assumption.
Qed.

Lemma eval_exprlist_correct:
  forall e le m el vl,
   eval_exprlist (Genv.globalenv p) e le m el vl ->
   forall m', Mem.tagmap stfun m = Some m' ->
              eval_exprlist (Genv.globalenv tp) e le m' el vl.
Proof.
  induction 1; constructor; eauto using eval_expr_correct.
Qed.

Section BLOCK_IS_VOLATILE.

Let block_is_volatile_preserved: forall b, 
  Events.block_is_volatile (Genv.globalenv tp) b = Events.block_is_volatile (Genv.globalenv p) b.
Proof.
  unfold Events.block_is_volatile. intros.
  functional inversion TRANSF_PROGRAM. subst. 
  rewrite Genv.find_var_info_transf. reflexivity.
Qed.

Let eventval_match_preserved:
  forall ev ty v,
    Events.eventval_match (Genv.globalenv p) ev ty v ->
    Events.eventval_match (Genv.globalenv tp) ev ty v.
Proof.
  functional inversion TRANSF_PROGRAM; subst.
  intros. eapply Events.eventval_match_preserved. intros. eapply Genv.find_symbol_transf. assumption.
Qed.

Let eventval_list_match_preserved:
  forall ev ty v,
    Events.eventval_list_match (Genv.globalenv p) ev ty v ->
    Events.eventval_list_match (Genv.globalenv tp) ev ty v.
Proof.
  induction 1; constructor; eauto.
Qed.

Lemma external_call_tagmap:
  forall ef vargs m t vres m',
    Events.external_call ef (Genv.globalenv p) vargs m t vres m' ->
    forall tm, Mem.tagmap stfun m = Some tm ->
               exists tm', Mem.tagmap stfun m' = Some tm' /\
                 Events.external_call ef (Genv.globalenv tp) vargs tm t vres tm'.
Proof.
  functional inversion TRANSF_PROGRAM; subst.
  destruct ef; simpl; inversion 1; subst.
   (* io *)
   esplit. split. eassumption. econstructor; eauto. 
   (* io *)
   esplit. split. eassumption. econstructor; eauto. 
   (* volatile load *)
   inversion H1; subst.
    esplit. split. eassumption. econstructor; eauto. econstructor; eauto. 
    rewrite Genv.find_symbol_transf. assumption.
   esplit. split. eassumption. econstructor; eauto. econstructor; eauto. erewrite Mem.tagmap_load; eauto. erewrite tagmap_option_tag_align; eauto.
   (* volatile store *)
   inversion H1; subst.
    esplit. split. eassumption. econstructor; eauto. econstructor; eauto. rewrite Genv.find_symbol_transf. assumption.
   intros.  exploit Mem.tagmap_store; eauto.
   destruct 1. destruct H5. esplit. split. eassumption.
   econstructor; eauto. econstructor; eauto. erewrite tagmap_option_tag_align; eauto.
   (* volatile load global *)
   inversion H2; subst.
    esplit. split. eassumption. econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption.
    econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption.
   esplit. split. eassumption. econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption. econstructor; eauto. erewrite Mem.tagmap_load; eauto. erewrite tagmap_option_tag_align; eauto.
   (* volatile store global *)
   inversion H2; subst.
    esplit. split. eassumption. econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption.
    econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption.
   intros.  exploit Mem.tagmap_store. eassumption. eassumption.
   destruct 1. destruct H6. esplit. split. eassumption.
   econstructor; eauto. rewrite Genv.find_symbol_transf. eassumption.  econstructor; eauto. erewrite tagmap_option_tag_align; eauto.
   (* malloc *)
   intros. exploit Mem.tagmap_alloc. eassumption. eassumption. reflexivity.   
   destruct 1. destruct H4.
   exploit Mem.tagmap_store. eassumption. eassumption.
   destruct 1.   destruct H6. esplit. split; eauto.
   econstructor; eauto.
   (* free *)   
   intros. exploit Mem.tagmap_free. eassumption. eassumption.
   destruct 1. destruct H5.
   esplit. split; eauto. econstructor; eauto. erewrite Mem.tagmap_load; eauto.
   exploit Mem.tagmap_tget_some. eexact H4. eassumption. destruct 1. destruct H7. inv H7. assumption.   
   (* memcpy *)
   intros; exploit Mem.tagmap_storebytes; eauto.
   destruct 1. destruct H10.
   esplit; split; eauto.
   econstructor; eauto.
   erewrite tagmap_option_tag_align; eauto.
   erewrite tagmap_option_tag_align; eauto.
   erewrite Mem.tagmap_loadbytes; eauto.
   (* annot *)
   intros. esplit; split; eauto. econstructor; eauto.
   (* annot val *)
   intros. esplit; split; eauto. econstructor; eauto.
   (* annot sem *)
   intros. esplit; split; eauto. econstructor; eauto.
Qed.

End BLOCK_IS_VOLATILE.

Lemma transf_step_correct:
  forall S1 t S2 (STEP: Csharpminor.step false (Genv.globalenv p) S1 t S2)
  T1 (MATCH: match_states S1 T1),
  (exists T2, step true (Genv.globalenv tp) T1 t T2 /\ match_states S2 T2).
Proof.
  inversion 1; subst; inversion 1; subst.

  (* skip seq *)
  inversion HCONT; subst. esplit. split. econstructor. constructor; eauto.

  (* skip block *)
  inversion HCONT; subst. esplit. split. econstructor. constructor; eauto.

  (* skip call *)
  exploit tagmap_region_free_aux.
  eassumption.
  eassumption.
  destruct 1. destruct H1.
  esplit. split.
  econstructor.
   destruct k; try contradiction; inv HCONT; trivial.
   unfold transf_function; simpl in *.
  rewrite compile_vars_region_free_aux. eassumption.
  constructor. assumption. assumption.

  (* expr *)
  esplit. split. econstructor. eapply eval_expr_correct. eassumption. assumption.
  constructor; auto.

  (* store *)
  exploit Mem.tagmap_store.
  eassumption.
  eassumption.
  destruct 1. destruct H2.
  esplit. split. econstructor.
   eapply eval_expr_correct. eassumption. assumption.
   eapply eval_expr_correct. eassumption. assumption.
   eassumption.
   assumption.
   erewrite tagmap_option_tag_align. eassumption. assumption.
   constructor; auto.

  (* call *)
  functional inversion TRANSF_PROGRAM.
  rewrite H2.
  esplit. split. econstructor.
  eapply eval_expr_correct. eassumption. assumption.
  eapply eval_exprlist_correct. eassumption. assumption.
  subst. eapply Genv.find_funct_transf. eassumption. 
  destruct fd; reflexivity.
  constructor; auto.
  constructor; auto.

  (* builtin *)
  exploit external_call_tagmap.
  eassumption.
  eassumption.
  destruct 1. destruct H1.
  esplit. split.
  econstructor. eapply eval_exprlist_correct. eassumption. assumption.
  eassumption.
  constructor; auto.

  (* seq *)
  esplit. split. constructor. constructor; auto. constructor; auto.

  (* if *)
  esplit. split. econstructor. eapply eval_expr_correct. eassumption. assumption. eassumption.
  constructor; auto.

  (* loop *)
  esplit. split. constructor. constructor; auto. constructor; auto.

  (* block *)
  esplit. split. constructor. constructor; auto. constructor; auto.

  (* exit seq *)
  inv HCONT.
  esplit. split. constructor. constructor; auto.

  (* exit 0 *)
  inv HCONT.
  esplit. split. constructor. constructor; auto.

  (* exit S *)
  inv HCONT.
  esplit. split. constructor. constructor; auto.

  (* switch *)
  esplit. split. constructor.
  eapply eval_expr_correct. eassumption. assumption.
  constructor; auto.

  (* return none *)
  exploit tagmap_region_free_aux.
  eassumption. eassumption.
  destruct 1. destruct H0.
  esplit. split. econstructor.
  unfold transf_function; simpl.
  rewrite compile_vars_region_free_aux. eassumption.
  constructor; auto. 
  apply match_call_cont; assumption.

  (* return some *)
  exploit tagmap_region_free_aux.
  eassumption. eassumption.
  destruct 1. destruct H1.
  esplit. split. econstructor.
  eapply eval_expr_correct. eassumption. assumption. 
  unfold transf_function; simpl.
  rewrite compile_vars_region_free_aux. eassumption.
  constructor; auto. 
  apply match_call_cont; assumption.

  (* label *)
  esplit. split. constructor. constructor; auto.

  (* goto *)
  destruct (match_find_label_aux) as [HLABEL _].
  generalize (HLABEL (fn_body f) lbl (call_cont k)); clear HLABEL.
  rewrite H.
  intro HLABEL.
  exploit HLABEL; clear HLABEL.
  eapply match_call_cont. eassumption.
  destruct 1. destruct H0.
  esplit. split. constructor. 
  eassumption.
  constructor; auto.

  (* call internal *)
  simpl.
  inversion H4; subst; try discriminate.
  exploit Mem.tagmap_alloc. eassumption. eassumption.
  simpl.
  unfold function_is_defined.
  rewrite H. rewrite H0.
  reflexivity.
  destruct 1. destruct H6.
  destruct (check_offsets_exists (fn_vars f)).
  
 
  esplit. split. econstructor.
  functional inversion TRANSF_PROGRAM; subst. rewrite Genv.find_symbol_transf. eassumption.
  functional inversion TRANSF_PROGRAM; subst. erewrite Genv.find_funct_ptr_transf. 
  2: eassumption. reflexivity.
  simpl. apply list_norepet_compile_vars. assumption.
  assumption.
  assumption.
  eassumption.
  eleft.
   reflexivity.
   eassumption.
   symmetry. erewrite Mem.tagmap_nextblock.  2: eassumption. reflexivity.
   reflexivity.
   reflexivity.
   simpl. f_equal.
   symmetry. apply stack_size_map.
   apply call_condition_ok.
   assumption.
   intros. generalize (ALIGN id sz al o). tauto.
   simpl. intros. unfold compile_vars, compile_vars' in H9. destruct (list_in_map_inv _ _ _ H9). destruct H10. destruct x1. destruct p1. destruct p1. inv H10. eauto.
  eassumption.
  simpl.
  unfold compile_vars, compile_vars'.
  rewrite assign_variables_aux_map.
  constructor.
   reflexivity.
   assumption.
   eapply tagmap_region_alloc.
   eassumption.
   assumption.   
   generalize (check_offsets_correct _ _ H8).
   intros.
   generalize (check_no_offsets_correct _ _ Hcheck).
   intro.
   unfold compile_vars in H9.
   congruence.
   assumption.

 (* external *)
 exploit external_call_tagmap.
 eassumption.
 eassumption.
 destruct 1. destruct H0.
 esplit. split. econstructor. eassumption. constructor; auto.

 (* return *)
 inv HCONT.
 esplit. split. econstructor. constructor; auto.

Qed.

Lemma transl_initial_states:
  forall S, Csharpminor.initial_state p S ->
  exists R, Csharpminor.initial_state tp R /\ match_states S R.
Proof.
  functional inversion TRANSF_PROGRAM.
  inversion 1; subst.
  esplit. split.
  econstructor.
  eapply Genv.init_mem_transf.
  eassumption.
  simpl. rewrite Genv.find_symbol_transf. eassumption.
  eapply Genv.find_funct_ptr_transf. eassumption.
  destruct f; auto.
  constructor.
   reflexivity.
   constructor.
  eapply Mem.tagmap_init_mem.
  eassumption.
Qed.

Lemma transl_final_states:
  forall S R r,
  match_states S R -> Csharpminor.final_state S r -> Csharpminor.final_state R r.
Proof.
  intros. inv H0. inv H. inv HCONT. constructor.
Qed.

Require Import Smallstep.

Theorem transl_program_correct:
  forward_simulation (Csharpminor.semantics false p) (Csharpminor.semantics true tp).
Proof.
  eapply forward_simulation_step; eauto.
  functional inversion TRANSF_PROGRAM.
  intro. eapply Genv.find_symbol_transf.
  eexact transl_initial_states.
  exact transl_final_states.
  exact transf_step_correct. 
Qed.


End MEM.
