Require Import Values.
Require Import Coqlib.
Require Import AST.
Require Import Integers.
Require Import Memory.


Module Mem.
Import Memtype.Mem.

Section WITHMEM.
Context `{M: Mem.MEM}.

Function region_alloc (b: block) (f: ident) (m: mem) (l: list (ident * (Z * Z * Z))) {struct l} : mem :=
  match l with
    | nil => m
    | (v, (sz, al, of)) :: l' =>
      match alloc m 0 sz (Tag_stack (Stack_var (Svar f v b sz al (Some of)))) with
        | (m', _) =>
          region_alloc b f m' l'
      end
  end.

Function region_free (b: block) (m: mem) (l: list (Z * Z * Z)) {struct l}: option mem :=
  match l with
    | nil => Some m
    | (sz, _, _) :: l' =>
      match free m b 0 sz with
        | Some m' => region_free (Zsucc b) m' l'
        | None => None
      end
  end.

Function alloc_dummyblocks (b: block) (f: ident) (m: mem) (l: list (ident * (Z * Z * Z))) {struct l} : mem :=
  match l with
    | nil => m
    | (v, (sz, al, of)) :: l' =>
      match alloc m 0 0 (Tag_stack (Stack_var (Svar f v b sz al (Some of)))) with
        | (m', _) =>
          alloc_dummyblocks b f m' l'
      end
  end.

Function region_size (l: list (Z * Z * Z)) {struct l}: Z :=
  match l with
    | nil => 0
    | (sz, _, _) :: l' => sz + region_size l'
  end.

Function tag_align (t: tag) {struct t}: Z :=
  match t with
    | Tag_stack (Stack_var sv) => stag_var_align sv
    | _ => 8
  end.

Function option_tag_align (t: option tag) {struct t}: Z :=
  match t with
    | Some t' => tag_align t'
    | _ => 8
  end.

Definition memval_strong_inject f v1 v2 :=
  memval_inject f v1 v2 /\ (v1 = Undef -> v2 = Undef).

Definition val_strong_inject f v1 v2 :=
  val_inject f v1 v2 /\ (v1 = Vundef -> v2 = Vundef).

(*
        forall (ALIGN: forall b' sz al of, 
                         tget m b = Some (Tag_stack (Stack_var b' sz al of)) ->
                         (align_chunk chunk | al)),
*)

Class StrongInject (valid_region_alloc: forall f: ident, list (ident * (Z * Z * Z)) -> Prop)
  : Type := {
   inject: meminj -> mem -> mem -> Prop;

   inject_func: forall f1 m m1, inject f1 m m1 -> forall f2 m2, inject f2 m m2 -> (f1, m1) = (f2, m2);

   inject_inj: forall f1 m1 m, inject f1 m1 m -> forall f2 m2, inject f2 m2 m -> (f1, m1) = (f2, m2);

(*
   inject_other: forall f m m',
                   inject f m m' ->
                   forall b, (forall b' fi i sz al of, tget m b <> Some (Tag_stack (Stack_var (Svar fi i b' sz al of)))) ->
                             forall r, f b = Some r -> r = (b, 0)
   ;

   inject_same: forall f m m',
                  inject f m m' ->
                  forall fi i b b' sz al of,
                    tget m b = Some (Tag_stack (Stack_var (Svar fi i b' sz al (Some of)))) ->
                    f b = Some (b', of)
                               ;
   
   tget_inject_same: forall f m m',
                       inject f m m' ->
                       forall fi i b b' sz al of,
                         tget m b = Some (Tag_stack (Stack_var (Svar fi i b' sz al (Some of)))) ->
                         exists sz', tget m' b' = Some (Tag_stack (Stack_main fi (Some sz')))
   ;

   tget_inject_var_main_size: forall f m m',
                       inject f m m' ->
                       forall fi i b b' sz al of,
                         tget m b = Some (Tag_stack (Stack_var (Svar fi i b' sz al (Some of)))) ->
                         forall sz',
                           tget m b' = Some (Tag_stack (Stack_main fi (Some sz'))) ->
                           of + sz <= sz'
   ;

   tget_inject_var_size: forall f m m',
                           inject f m m' ->
                           forall fi i b b' sz al of,
                           tget m b = Some (Tag_stack (Stack_var (Svar fi i b' sz al (Some of)))) ->
                           0 <= sz <= Int.max_unsigned  /\ 0 < al /\ (al | sz) /\ 0 <= of
   ;

   tget_inject_main_size: forall f m m',
                            inject f m m' ->
                            forall b f' sz',
                              tget m b = Some (Tag_stack (Stack_main f' (Some sz'))) ->
                              0 <= sz' <= Int.max_unsigned
   ;
*)
      
   tget_inject: forall f m1 m2,
                  inject f m1 m2 ->
  forall b,
    tget m1 b = tget m2 b
   ;

   tget_inject_align_block: forall f m1 m2,
                        inject f m1 m2 ->
                        forall b b' o, f b = Some (b', o) ->
                          (option_tag_align (Mem.tget m1 b) | option_tag_align (Mem.tget m1 b')) ;

   tget_inject_align_offset: forall f m1 m2,
                        inject f m1 m2 ->
                        forall b b' o, f b = Some (b', o) ->
                                       (option_tag_align (Mem.tget m1 b) | o)
   ;
   
   tget_inject_heap_freestore: forall f m1 m2,
                           inject f m1 m2 ->
                           forall b b' o, f b = Some (b', o) ->
                                          Mem.tget m1 b = Some (Mem.Tag_heap Mem.Heap_freestore) ->
                                          Mem.tget m1 b' = Some (Mem.Tag_heap Mem.Heap_freestore)
   ;

   mi_freeblocks f m1 m2:
     inject f m1 m2 ->
     forall b, ~(valid_block m1 b) -> f b = None
   ;

   valid_block_inject_1 f m1 m2 b1 b2 delta:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     valid_block m1 b1
   ;

   valid_block_inject_2 f m1 m2 b1 b2 delta:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     valid_block m2 b2
   ;

   perm_inject f m1 m2 b1 b2 delta ofs k p:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->
     perm m1 b1 ofs k p -> perm m2 b2 (ofs + delta) k p
   ;

   range_perm_inject:
     forall f m1 m2 b1 b2 delta lo hi k p,
       f b1 = Some(b2, delta) ->
       inject f m1 m2 ->
       range_perm m1 b1 lo hi k p -> range_perm m2 b2 (lo + delta) (hi + delta) k p
   ;

   valid_access_inject f m1 m2 chunk b1 ofs b2 delta p:
     f b1 = Some(b2, delta) ->
     inject f m1 m2 ->     
     (align_chunk chunk | option_tag_align (tget m1 b1)) ->
     valid_access m1 chunk b1 ofs p ->
  valid_access m2 chunk b2 (ofs + delta) p
   ;

valid_pointer_inject f m1 m2 b1 ofs b2 delta:
  f b1 = Some(b2, delta) ->
  inject f m1 m2 ->
  valid_pointer m1 b1 ofs = true ->
  valid_pointer m2 b2 (ofs + delta) = true
   ;

weak_valid_pointer_inject f m1 m2 b1 ofs b2 delta:
  f b1 = Some(b2, delta) ->
  inject f m1 m2 ->
  weak_valid_pointer m1 b1 ofs = true ->
  weak_valid_pointer m2 b2 (ofs + delta) = true
   ;

address_inject f m1 m2 b1 ofs1 b2 delta p:
  inject f m1 m2 ->
  perm m1 b1 (Int.unsigned ofs1) Cur p ->
  f b1 = Some (b2, delta) ->
  Int.unsigned (Int.add ofs1 (Int.repr delta)) = Int.unsigned ofs1 + delta
   ;

valid_pointer_inject_no_overflow f m1 m2 b ofs b' delta:
  inject f m1 m2 ->
  valid_pointer m1 b (Int.unsigned ofs) = true ->
  f b = Some(b', delta) ->
  0 <= Int.unsigned ofs + Int.unsigned (Int.repr delta) <= Int.max_unsigned
   ;

weak_valid_pointer_inject_no_overflow f m1 m2 b ofs b' delta:
  inject f m1 m2 ->
  weak_valid_pointer m1 b (Int.unsigned ofs) = true ->
  f b = Some(b', delta) ->
  0 <= Int.unsigned ofs + Int.unsigned (Int.repr delta) <= Int.max_unsigned
   ;

valid_pointer_inject_val f m1 m2 b ofs b' ofs':
  inject f m1 m2 ->
  valid_pointer m1 b (Int.unsigned ofs) = true ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  valid_pointer m2 b' (Int.unsigned ofs') = true
   ;

weak_valid_pointer_inject_val f m1 m2 b ofs b' ofs':
  inject f m1 m2 ->
  weak_valid_pointer m1 b (Int.unsigned ofs) = true ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  weak_valid_pointer m2 b' (Int.unsigned ofs') = true
   ;

inject_no_overlap f m1 m2 b1 b2 b1' b2' delta1 delta2 ofs1 ofs2:
  inject f m1 m2 ->
  b1 <> b2 ->
  f b1 = Some (b1', delta1) ->
  f b2 = Some (b2', delta2) ->
  perm m1 b1 ofs1 Max Nonempty ->
  perm m1 b2 ofs2 Max Nonempty ->
  b1' <> b2' \/ ofs1 + delta1 <> ofs2 + delta2
   ;

different_pointers_inject f m m' b1 ofs1 b2 ofs2 b1' delta1 b2' delta2:
  inject f m m' ->
  b1 <> b2 ->
  valid_pointer m b1 (Int.unsigned ofs1) = true ->
  valid_pointer m b2 (Int.unsigned ofs2) = true ->
  f b1 = Some (b1', delta1) ->
  f b2 = Some (b2', delta2) ->
  b1' <> b2' \/
  Int.unsigned (Int.add ofs1 (Int.repr delta1)) <>
  Int.unsigned (Int.add ofs2 (Int.repr delta2))
   ;

disjoint_or_equal_inject:
  forall f m m' b1 b1' delta1 b2 b2' delta2 ofs1 ofs2 sz,
  inject f m m' ->
  f b1 = Some(b1', delta1) ->
  f b2 = Some(b2', delta2) ->
  range_perm m b1 ofs1 (ofs1 + sz) Max Nonempty ->
  range_perm m b2 ofs2 (ofs2 + sz) Max Nonempty ->
  sz > 0 ->
  b1 <> b2 \/ ofs1 = ofs2 \/ ofs1 + sz <= ofs2 \/ ofs2 + sz <= ofs1 ->
  b1' <> b2' \/ ofs1 + delta1 = ofs2 + delta2
             \/ ofs1 + delta1 + sz <= ofs2 + delta2 
             \/ ofs2 + delta2 + sz <= ofs1 + delta1
   ;

aligned_area_inject:
  forall f m m' b ofs al sz b' delta,
  inject f m m' ->
  al = 1 \/ al = 2 \/ al = 4 \/ al = 8 -> sz > 0 ->
  (al | sz) ->
  range_perm m b ofs (ofs + sz) Cur Nonempty ->
  (al | ofs) ->
  f b = Some(b', delta) ->
  (al | ofs + delta)
   ;

load_inject f m1 m2 chunk b1 ofs b2 delta v1:
  inject f m1 m2 ->
  load chunk m1 b1 ofs = Some v1 ->
  f b1 = Some (b2, delta) ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists v2, load chunk m2 b2 (ofs + delta) = Some v2 /\ val_strong_inject f v1 v2
;

loadv_inject f m1 m2 chunk b1 o1 a2 v1:
  inject f m1 m2 ->
  loadv chunk m1 (Vptr b1 o1) = Some v1 ->
  val_inject f (Vptr b1 o1) a2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists v2, loadv chunk m2 a2 = Some v2 /\ val_strong_inject f v1 v2
;

loadbytes_inject f m1 m2 b1 ofs len b2 delta bytes1:
  inject f m1 m2 ->
  loadbytes m1 b1 ofs len = Some bytes1 ->
  f b1 = Some (b2, delta) ->
  exists bytes2, loadbytes m2 b2 (ofs + delta) len = Some bytes2
              /\ list_forall2 (memval_strong_inject f) bytes1 bytes2
;
store_mapped_inject f chunk m1 b1 ofs v1 n1 m2 b2 delta v2:
  inject f m1 m2 ->
  store chunk m1 b1 ofs v1 = Some n1 ->
  f b1 = Some (b2, delta) ->
  val_strong_inject f v1 v2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists n2,
    store chunk m2 b2 (ofs + delta) v2 = Some n2
    /\ inject f n1 n2
;
storev_mapped_inject f chunk m1 b1 o1 v1 n1 m2 a2 v2:
  inject f m1 m2 ->
  storev chunk m1 (Vptr b1 o1) v1 = Some n1 ->
  val_inject f (Vptr b1 o1) a2 ->
  val_strong_inject f v1 v2 ->
  (align_chunk chunk | option_tag_align (tget m1 b1)) ->
  exists n2,
    storev chunk m2 a2 v2 = Some n2 /\ inject f n1 n2
;
storebytes_mapped_inject f m1 b1 ofs bytes1 n1 m2 b2 delta bytes2:
  inject f m1 m2 ->
  storebytes m1 b1 ofs bytes1 = Some n1 ->
  f b1 = Some (b2, delta) ->
  list_forall2 (memval_strong_inject f) bytes1 bytes2 ->
  exists n2,
    storebytes m2 b2 (ofs + delta) bytes2 = Some n2
    /\ inject f n1 n2
;
alloc_parallel_inject:
  forall f m1 m2 lo1 hi1 t m1' b1 lo2 hi2,
  inject f m1 m2 ->
  alloc m1 lo1 hi1 t = (m1', b1) ->
  lo2 <= lo1 -> hi1 <= hi2 ->
  (forall st, t <> Tag_stack st) ->
  exists f', exists m2', exists b2,
  alloc m2 lo2 hi2 t = (m2', b2)
  /\ inject f' m1' m2'
  /\ inject_incr f f'
  /\ f' b1 = Some(b2, 0)
  /\ (forall b, b <> b1 -> f' b = f b)

;
alloc_region_inject:
  forall f m1 m2,
    inject f m1 m2 ->
    forall m1' sz fi b1,
      alloc m1 0 0 (Tag_stack (Stack_main fi (Some sz))) = (m1', b1) ->
      forall l,
      valid_region_alloc fi l ->
      (forall id1 sz1 al1 of1, 
         In (id1, (sz1, al1, of1)) l ->
         (0 <= sz1 /\ 0 < al1 /\ (al1 | sz1) /\ (al1 | of1) /\ (al1 | 8) /\ 0 <= of1 /\ of1 + sz1 <= sz)) ->
      (forall n1 id1 sz1 al1 of1, 
         List.nth_error l n1 = Some (id1, (sz1, al1, of1)) ->
         forall n2 id2 sz2 al2 of2,
           List.nth_error l n2 = Some (id2, (sz2, al2, of2)) ->
           n1 <> n2 ->
           (of1 + sz1 <= of2 \/ of2 + sz2 <= of1)) ->
      sz = region_size (map snd l) ->
      exists f', exists m2', exists b2,
        alloc m2 0 sz (Tag_stack (Stack_main fi (Some sz))) = (m2', b2) /\
        inject f' (region_alloc b1 fi m1' l) (alloc_dummyblocks b2 fi m2' l) /\
        inject_incr f f' /\
        f' b1 = None /\
        (forall n i sz al of, List.nth_error l n = Some (i, (sz, al, of)) ->
                   f' (b1 + 1 + Z.of_nat n) = Some (b2, of)) /\
        forall b, ~ (b1 + 1 <= b < b1 + 1 + Z.of_nat (length l)) ->
                  f' b = f b
;

free_parallel_inject:
  forall f m1 m2,
  inject f m1 m2 ->
  forall b1 b2 o, f b1 = Some (b2, o) ->
                  forall lo hi m1',
                  free m1 b1 lo hi = Some m1' ->
                  exists m2',
                    free m2 b2 (lo + o) (hi + o) = Some m2' /\
                    inject f m1' m2'
;

free_region_inject:
  forall f m1 m2,
    inject f m1 m2 ->
    forall fi l,
      valid_region_alloc fi l ->
      forall b1 sz,
          tget m1 b1 = Some (Tag_stack (Stack_main fi (Some sz))) ->
          (forall n i sz al of,
             List.nth_error l n = Some (i, (sz, al, of)) ->
             tget m1 (b1 + 1 + Z.of_nat n) = Some (Tag_stack (Stack_var (Svar fi i b1 sz al (Some of))))) ->
          forall m1',
            region_free (b1 + 1) m1 (map snd l) = Some m1' ->
            exists m2',
              free m2 b1 0 sz = Some m2' /\
              inject f m1' m2'
;

inject_nextblock:
  forall f m1 m2,
    inject f m1 m2 ->
    nextblock m2 = nextblock m1
;

neutral_inject m:
  inject_neutral (nextblock m) m ->
  inject (flat_inj (nextblock m)) m m

}.

End WITHMEM.
End Mem.
