Require Import Memimpl Tagmap.
Require Import Values.
Require Import Coqlib.

Section TMap.

Import Mem.

Variable (f: stag -> option stag). (* we cannot change heap tags due to empty mem *)

Lemma check_tmap (t: block -> option tag) (n: block):
  {forall b, 0 < b < n -> forall ta, t b = Some ta -> tagf f ta <> None} +
  {~ forall b, 0 < b < n -> forall ta, t b = Some ta -> tagf f ta <> None}.
Proof.
  destruct (zle n 0).
   left. intros; omega.
  assert (0 <= n) by omega.
  revert g. pattern n. apply natlike_rec.
   intro; omega.
  intros.
  destruct (zeq x 0).
   subst. left. intros. omega.
  assert (x > 0) by omega.
  clear g.
  destruct H1.
   assumption.
   case_eq (t x).
      intros. case_eq (tagf f t0).
       intros. left. intros. destruct (zeq b x).
        congruence.
       eapply n1. 2: eassumption. omega.
      intros. right. intro. eapply H4.
       2: eassumption. omega.
       assumption.
    intros. left. intros. destruct (zeq b x).
     congruence.
    eapply  n1. 2: eassumption. omega.
   right. intro. eapply n1. intros. eapply H1. 2: eassumption. omega.
  assumption.
Defined.

Function tmap t (b: block) := match t b with None => None | Some t' => tagf f t' end.

(*
Lemma tmap_nextblock_notget:
  forall n, (n > 0) ->
            forall t,
              (forall b, 0 < b < n -> forall ta, t b = Some ta -> tagf f ta <> None) ->
              (forall b, b >= n <-> t b = None) ->
              forall 
*)

Program Definition tagmap (m: mem): option mem :=
  if check_tmap (tget m) (nextblock m)
  then Some
         (mkmem m.(mem_contents) m.(mem_access) (nextblock m) (nextblock_pos m)
                                                                   m.(access_max) m.(nextblock_nocontents) m.(nextblock_noaccess) (tmap (tget m)) _ _ m.(block_bounds) m.(nextblock_nobounds) m.(block_bounds_perm))
  else None.
Next Obligation.
  unfold tmap. case_eq (Memimpl.tget m b).
   intros.
    destruct (zle b 0).
     exploit (nextblock_default m). instantiate (1 := b). omega.
     intro. rewrite H0 in H1. inv H1. simpl. generalize (nextblock_pos m). intro.
     split; try discriminate; unfold block in *; simpl in *; omega.
    destruct (zlt b (Memimpl.nextblock m)).
    apply H in H0.
    split; try omega; congruence.
    omega.
   apply nextblock_notget in g0. congruence.
  intro. apply nextblock_notget in H0. tauto.
Qed.
Next Obligation.
  apply (nextblock_default m) in H0. unfold tmap. rewrite H0. reflexivity.
Qed.


Lemma tagmap_empty: tagmap empty = Some empty.
Proof.
  unfold tagmap; simpl.
   destruct (check_tmap (fun b => if zlt b 1 then Some (Tag_heap Heap_global) else None) 1).
    unfold Memimpl.empty. f_equal. apply mkmem_ext; simpl; try reflexivity. simpl.
    apply FunctionalExtensionality.functional_extensionality. unfold tmap.
     intro. destruct (zlt x 1); reflexivity.
    destruct n. intros; omega.
Qed.

Lemma tagmap_nextblock: forall m m' : mem, tagmap m = Some m' -> nextblock m' = nextblock m.
Proof.
 unfold tagmap. intros until m'. destruct (check_tmap (tget m) (nextblock m)); simpl in *.
  intro. inv H. reflexivity.
  discriminate.
Qed.


Lemma tagmap_load: forall m m' : mem,
   tagmap m = Some m' ->
   forall (chunk : AST.memory_chunk) (b : block) (o : Z),
   load chunk m' b o = load chunk m b o.
Proof.
  unfold tagmap. intros until m'. destruct (check_tmap (tget m) (nextblock m)); simpl in *.
  intro. inv H. unfold Memimpl.load. simpl. reflexivity.
  discriminate.
Qed.

Lemma tagmap_loadv:
  forall m m' : mem,
    tagmap m = Some m' ->
    forall (chunk : AST.memory_chunk) (v : val),
      loadv chunk m' v = loadv chunk m v.
Proof.
  unfold loadv. destruct v; try reflexivity. eauto using tagmap_load.
Qed.

Lemma tagmap_loadbytes:
 forall m m' : mem,
   tagmap m = Some m' ->
   forall (b : block) (o l : Z), loadbytes m' b o l = loadbytes m b o l.
Proof.
  unfold tagmap. intros until m'. destruct (check_tmap (tget m) (nextblock m)); simpl in *.
  intro. inv H. reflexivity.
  discriminate.
Qed.

Lemma tagmap_tget_none: forall m m' : mem,
   tagmap m = Some m' ->
   forall b : block, tget m b = None -> tget m' b = None.
Proof.
  unfold tagmap. intros until m'. destruct (check_tmap (tget m) (nextblock m)); simpl in *.
  intro. inv H. simpl. intros. unfold tmap. rewrite H. reflexivity.
  discriminate.
Qed.


Lemma tagmap_tget_some:
forall m m' : mem,
   tagmap m = Some m' ->
   forall (t : tag) (b : block),
   tget m b = Some t ->
   exists t' : tag, tagf f t = Some t' /\ tget m' b = Some t'.
Proof.
  unfold tagmap. intros until m'. destruct (check_tmap (tget m) (nextblock m)); simpl in *.
  intro. inv H. simpl. unfold tmap.
  intros. rewrite H.
  destruct (zle b 0).
   exploit (nextblock_default m). instantiate (1 := b). omega. intro. rewrite H in H0. inv H0. simpl. eauto.
  destruct (zlt b (Memimpl.nextblock m)).
   refine (_ (n b _ _ H)).
   destruct (tagf f t); try congruence. eauto.
   omega.
  apply nextblock_notget in g0. congruence.
 discriminate.
Qed.

Lemma tagmap_store:
 forall (m : mem) (chunk : AST.memory_chunk) (b : block) 
     (o : Z) (v : val) (m' : mem),
   store chunk m b o v = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ store chunk tm b o v = Some tm'.
Proof.
 cut (forall (m : mem) (chunk : AST.memory_chunk) (b : block) 
     (o : Z) (v : val) (m' : mem),
   store chunk m b o v = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ forall tm'', tm'' = tm -> store chunk tm'' b o v = Some tm'
   ).
  intros. exploit H; eauto. destruct 1. destruct H2. eauto.
 unfold tagmap. intros. revert  H0. case_eq (check_tmap (tget m) (nextblock m)); try discriminate.
 intros. inv H1.
 unfold store in H; simpl in H. unfold Memimpl.store in H.
 destruct (valid_access_dec m chunk b o Writable); try discriminate.
 inv H.
 simpl. simpl in H0. rewrite H0. simpl.
 esplit. split. reflexivity. 
 intros. unfold Memimpl.store. destruct (valid_access_dec tm'' chunk b o Writable).
  subst. simpl in *. f_equal. apply mkmem_ext; reflexivity.
 exfalso. unfold Memimpl.valid_access, Memimpl.range_perm, Memimpl.perm in *. subst. contradiction.
Qed.

Lemma tagmap_storev: forall (m : mem) (chunk : AST.memory_chunk) (a v : val) (m' : mem),
   storev chunk m a v = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ storev chunk tm a v = Some tm'.
Proof.
  unfold storev. destruct a; try discriminate.
  eauto using tagmap_store.
Qed.

Lemma tagmap_storebytes:
  forall (m : mem) (b : block) (o : Z) (v : list memval) (m' : mem),
   storebytes m b o v = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ storebytes tm b o v = Some tm'.
Proof.
  cut (forall (m : mem) (b : block) (o : Z) (v : list memval) (m' : mem),
   storebytes m b o v = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ forall tm'', tm'' = tm -> storebytes tm'' b o v = Some tm').
  intros. exploit H; eauto. destruct 1. destruct H2. eauto.
 unfold tagmap. intros. revert  H0. case_eq (check_tmap (tget m) (nextblock m)); try discriminate.
 intros. inv H1.
 unfold storebytes in H; simpl in H. unfold Memimpl.storebytes in H.
 destruct (range_perm_dec m b o (o + Z.of_nat (length v)) Cur Writable); try discriminate.
 inv H.
 simpl. simpl in H0. rewrite H0. simpl.
 esplit. split. reflexivity. 
 intros. unfold Memimpl.storebytes. destruct (range_perm_dec tm'' b o (o + Z.of_nat (length v)) Cur Writable).
  subst. simpl in *. f_equal. apply mkmem_ext; reflexivity.
 exfalso. unfold Memimpl.range_perm, Memimpl.perm in *. subst. contradiction.
Qed.  

Lemma tagmap_alloc:
forall (m : mem) (lo hi : Z) (t : tag) (b : block) (m' : mem),
   alloc m lo hi t = (m', b) ->
   forall tm : mem,
   tagmap m = Some tm ->
   forall t' : tag,
   tagf f t = Some t' ->
   exists tm' : mem, tagmap m' = Some tm' /\ alloc tm lo hi t' = (tm', b).
Proof.
  unfold alloc; simpl. unfold talloc.
  intros until m'.
  destruct (talloc_strong m lo hi t).
   destruct s. destruct e. destruct H. destruct H0. destruct H1. destruct H2. destruct H3.
   intro. inv H5.
   intros.
   destruct (talloc_strong tm lo hi t').
   destruct s. destruct e. destruct H7. destruct H8. destruct H9. destruct H10. destruct H11.
   unfold tagmap in H5.
   destruct (check_tmap (tget m) (nextblock m)); try discriminate.
   inv H5.
   inv H7. simpl in *.
   destruct x; simpl in *; subst.
   inv H. simpl in *.
   destruct m'; simpl in *; subst. 
    unfold tagmap. simpl.
    destruct (
 check_tmap
         (fun b' : block =>
          if zeq b' (Memimpl.nextblock m)
          then Some t
          else
           if zeq b' (Memimpl.nextblock m)
           then Some (Tag_heap Heap_global)
           else Memimpl.tget m b') (Z.succ (Memimpl.nextblock m))
      ).
     esplit. split. reflexivity. f_equal. apply mkmem_ext; try reflexivity.
     apply FunctionalExtensionality.functional_extensionality. unfold tmap. simpl. intro.
     destruct (zeq x (Memimpl.nextblock m)).
      subst. congruence.
     reflexivity.
   exfalso. revert n H6 n0. clear.
   intros. destruct n0.
    intros. destruct (zeq b (Memimpl.nextblock m)).
     subst. congruence.
    eapply n. 2: eassumption. omega.
Qed.

Lemma tagmap_free: forall (m : mem) (b : block) (lo hi : Z) (m' : mem),
   free m b lo hi = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem, tagmap m' = Some tm' /\ free tm b lo hi = Some tm'.
Proof.
  unfold free; simpl. unfold Memimpl.free. intros.
  destruct (range_perm_dec m b lo hi Cur Freeable); try discriminate.
  destruct (range_perm_dec tm b lo hi Cur Freeable).
   unfold tagmap in H0. revert H0.
   case_eq (check_tmap (tget m) (nextblock m)).
    intros. inv H1.
    inv H. unfold unchecked_free. unfold tagmap. simpl.
    unfold tget, nextblock in H0. simpl in H0. rewrite H0.
    esplit. split. reflexivity. f_equal. apply mkmem_ext; reflexivity.
  discriminate.
 exfalso.  unfold Memimpl.range_perm, Memimpl.perm in *. revert H0.
 unfold tagmap. destruct (check_tmap (tget m) (nextblock m)); try discriminate. intro. inv H0. contradiction.
Qed.
  
Lemma tagmap_perm:
  forall m m' : mem,
   tagmap m = Some m' ->
   forall (b : block) (o : Z) (k : perm_kind) (p : permission),
   perm m' b o k p <-> perm m b o k p.
Proof.
  unfold tagmap. intros.
  destruct (check_tmap (tget m) (nextblock m)); try discriminate. inv H. simpl. unfold Memimpl.perm. simpl. tauto.
Qed.

Lemma tagmap_drop:
forall (m : mem) (b : block) (lo hi : Z) (p : permission) (m' : mem),
   drop_perm m b lo hi p = Some m' ->
   forall tm : mem,
   tagmap m = Some tm ->
   exists tm' : mem,
     tagmap m' = Some tm' /\ drop_perm tm b lo hi p = Some tm'.
Proof.
  unfold drop_perm; simpl. unfold Memimpl.drop_perm. intros.
  destruct (range_perm_dec m b lo hi Cur Freeable); try discriminate.
  destruct (range_perm_dec tm b lo hi Cur Freeable).
   unfold tagmap in H0. revert H0.
   case_eq (check_tmap (tget m) (nextblock m)).
    intros. inv H1.
    inv H. unfold tagmap. simpl.
    unfold tget, nextblock in H0. simpl in H0. rewrite H0.
    esplit. split. reflexivity. f_equal. apply mkmem_ext; reflexivity.
  discriminate.
 exfalso.  unfold Memimpl.range_perm, Memimpl.perm in *. revert H0.
 unfold tagmap. destruct (check_tmap (tget m) (nextblock m)); try discriminate. intro. inv H0. contradiction.
Qed.

Lemma tagmap_inj:
  (forall s1 t : stag,
    f s1 = Some t -> forall s2 : stag, f s2 = Some t -> s1 = s2) ->
   forall m1 tm : mem,
   tagmap m1 = Some tm -> forall m2 : mem, tagmap m2 = Some tm -> m1 = m2.
Proof.
  unfold tagmap. intros.
  destruct (check_tmap (tget m1) (nextblock m1)); try discriminate.
  destruct (check_tmap (tget m2) (nextblock m2)); try discriminate.
  inv H0. inv H1.
  destruct m1; destruct m2; simpl in *.
  apply mkmem_ext; eauto.
  subst.
  revert H H5 n n0 nextblock_notget nextblock_notget0 nextblock_pos0 nextblock_default nextblock_default0. clear.
  intros.
  apply FunctionalExtensionality.functional_extensionality.
  intros. assert (tmap tget1 x = tmap tget0 x) by congruence.
  revert H0. unfold tmap.
  case_eq (tget0 x).
   intros until 1.
   case_eq (tget1 x).
    intros until 1.
    destruct (zlt 0 x).
     destruct (zlt x (nextblock0)).
      generalize (n _ (conj l l0) _ H0).
      generalize (n0 _ (conj l l0) _ H1).
      unfold tagf. destruct t0; destruct t; simpl; try congruence.
      intros _. case_eq (f s); congruence.
      case_eq (f s); congruence.
      case_eq (f s); case_eq (f s0); try congruence. intros. inv H7. f_equal. f_equal. eauto.
     apply nextblock_notget in g. congruence.
    exploit nextblock_default. instantiate (1 := x). omega. intro. rewrite H0 in H2. inv H2.
    exploit nextblock_default0. instantiate (1 := x). omega. intro. rewrite H1 in H2. inv H2.
    reflexivity.
   intro. apply nextblock_notget0 in H1. apply nextblock_notget in H1. congruence.
  intro. case_eq (tget1 x); try congruence.
  intros. apply nextblock_notget in H0. apply nextblock_notget0 in H0. congruence.
Qed.

Instance Tagmap : Mem.TAGMAP f := 
  {
    tagmap := tagmap
  }.
Proof.
  apply tagmap_empty.
  apply tagmap_nextblock.
  apply tagmap_load.
  apply tagmap_loadv.
  apply tagmap_loadbytes.
  apply tagmap_tget_none.
  apply tagmap_tget_some.
  apply tagmap_store.
  apply tagmap_storev.
  apply tagmap_storebytes.
  apply tagmap_alloc.
  apply tagmap_free.
  apply tagmap_perm.
  apply tagmap_drop.
  apply tagmap_inj.
Qed.

End TMap.
